import React, {useEffect, useRef} from 'react';
import {View, Animated, ListViewDataSource} from 'react-native';
import PropTypes from 'prop-types';
import {width, height} from './utils';

const KEY = '__PARALLAX_SCROLL__';
const RATIO = 9 / 16;

const ParallaxScroll = (props) => {
  const scrollY = useRef(new Animated.Value(0)).current;
  let isHeaderVisibible = true;
  const {
    data,
    style: wrapperStyle,
    width,
    height,
    sections,
    children,
    renderRow,
    renderItem,
    dataSource,
    scrollStyle,
    renderHeader,
    useNativeDriver,
    scrollableComponent,
    contentContainerStyle,
    renderParallaxBackground,
    renderParallaxForeground,
    ...scrollViewProps
  } = props;

  const _onAnimatedScroll = Animated.event(
    [{nativeEvent: {contentOffset: {y: scrollY}}}],
    {
      listener: _onScroll,
      useNativeDriver: props.useNativeDriver,
    },
  );

  useEffect(() => {
    if (Animated.ScrollView !== props.scrollableComponent) {
      scrollableComponent = Animated.createAnimatedComponent(
        props.scrollableComponent,
      );
    }
  }, []);

  const style = [scrollStyle, {width, height}];
  const stickyHeaderIndices = [];
  const ScrollableComponent = scrollableComponent;
  const isRenderChildComponents =
    !(data && renderItem) &&
    !((sections && renderItem) || (sections && sections[0].renderItem)) &&
    !(dataSource && renderRow);

  if (isRenderChildComponents && renderParallaxForeground) {
    stickyHeaderIndices.push(stickyHeaderIndices.length);
  }

  const _getFlatData = (data) => {
    return [{key: KEY}, ...data];
  };

  const _getDataSource = (dataSource) => {
    return dataSource.cloneWithRowsAndSections({
      [KEY]: [''],
      ...dataSource._dataBlob,
    });
  };

  const _getSectionData = (sections) => {
    if (props.renderItem) {
      return [{data: [{key: KEY}], key: KEY}, ...sections];
    }

    return [
      {
        data: [{key: KEY}],
        key: KEY,
        renderItem: _renderBackgroundPlaceholder,
      },
      ...sections,
    ];
  };

  const _onScroll = (e) => {
    const contentOffsetY = e.nativeEvent.contentOffset.y;
    const {
      onScroll,
      renderHeader,
      headerHeight,
      isHeaderFixed,
      onHeaderFixed,
      parallaxHeight,
      headerFixedTransformY,
      onChangeHeaderVisibility,
    } = props;

    const isHeaderFixedAfterScroll =
      contentOffsetY > parallaxHeight - headerHeight + headerFixedTransformY;
    const isHeaderVisibibleAfterScroll = contentOffsetY < parallaxHeight;

    if (onScroll) {
      onScroll(e);
    }

    if (
      renderHeader &&
      isHeaderFixed &&
      isHeaderFixed !== isHeaderFixedAfterScroll
    ) {
      isHeaderFixed = isHeaderFixedAfterScroll;
      onHeaderFixed(isHeaderFixedAfterScroll);
    }

    if (
      renderHeader &&
      !isHeaderFixed &&
      isHeaderVisibible !== isHeaderVisibibleAfterScroll
    ) {
      isHeaderVisibible = isHeaderVisibibleAfterScroll;
      onChangeHeaderVisibility(isHeaderVisibibleAfterScroll);
    }
  };

  const _renderRow = (rowData, sectionID, rowID, highlightRow) => {
    if (sectionID === KEY) {
      return _renderBackgroundPlaceholder();
    }

    return props.renderRow(rowData, sectionID, rowID, highlightRow);
  };

  const _renderItem = (e) => {
    if (e.item.key === KEY) {
      return _renderBackgroundPlaceholder();
    }

    return props.renderItem(e);
  };

  const _renderBackgroundPlaceholder = () => {
    const {parallaxHeight, renderBackgroundPlaceholder} = props;

    if (renderBackgroundPlaceholder) {
      return renderBackgroundPlaceholder({
        animatedValue: scrollY,
        height: parallaxHeight,
      });
    }

    return <View style={{height: parallaxHeight}} />;
  };

  const _ref = (ref) => {
    if (typeof props.innerRef === 'function' && ref && ref._component) {
      props.innerRef(ref._component);
    }
  };

  const _renderParallaxBackground = () => {
    const {
      width,
      parallaxHeight: height,
      isBackgroundScalable,
      renderParallaxBackground,
      fadeOutParallaxBackground,
      parallaxBackgroundScrollSpeed,
      backgroundScale,
      backgroundScaleOrigin,
    } = props;

    const topOrigin = backgroundScaleOrigin === 'top';

    const translateY = !height
      ? 0
      : scrollY.interpolate({
          inputRange: [...(topOrigin ? [-height, 0] : [0]), height],
          outputRange: [
            ...(topOrigin ? [-(height / backgroundScale) + height, 0] : [0]),
            -(height / parallaxBackgroundScrollSpeed),
          ],
          extrapolateLeft: 'extend',
          extrapolateRight: 'extend',
        });

    const scale =
      !isBackgroundScalable || !height
        ? 1
        : scrollY.interpolate({
            inputRange: [-height, 0],
            outputRange: [backgroundScale, 1],
            extrapolateLeft: 'extend',
            extrapolateRight: 'clamp',
          });

    const opacity =
      !fadeOutParallaxBackground || !height
        ? 1
        : scrollY.interpolate({
            inputRange: [0, height],
            outputRange: [1, 0],
            extrapolate: 'clamp',
          });

    return (
      <Animated.View
        style={{
          position: 'absolute',
          width,
          height,
          opacity,
          transform: [{translateY}, {scale}],
        }}
        pointerEvents="box-none">
        {renderParallaxBackground({
          width,
          height,
          animatedValue: scrollY,
        })}
      </Animated.View>
    );
  };

  const _renderParallaxForeground = () => {
    const {
      width,
      parallaxHeight: height,
      renderParallaxForeground,
      fadeOutParallaxForeground,
      parallaxForegroundScrollSpeed,
    } = props;

    /* eslint-disable indent */
    const translateY = !height
      ? 1
      : scrollY.interpolate({
          inputRange: [0, height],
          outputRange: [0, -(height / parallaxForegroundScrollSpeed)],
          extrapolateRight: 'extend',
          extrapolateLeft: 'clamp',
        });

    const opacity =
      !fadeOutParallaxForeground || !height
        ? 1
        : scrollY.interpolate({
            inputRange: [0, height],
            outputRange: [1, 0],
            extrapolateRight: 'extend',
            extrapolateLeft: 'clamp',
          });
    /* eslint-disable indent */

    const wrapperStyle = {
      position: 'absolute',
      top: 0,
      width,
      height,
      zIndex: 1,
    };

    return (
      <View style={wrapperStyle} pointerEvents="box-none">
        <Animated.View
          style={{
            position: 'absolute',
            width,
            height,
            opacity,
            transform: [{translateY}],
          }}
          pointerEvents="box-none">
          {renderParallaxForeground({
            width,
            height,
            animatedValue: scrollY,
          })}
        </Animated.View>
      </View>
    );
  };

  const _renderHeader = () => {
    const {
      width,
      headerHeight: height,
      renderHeader,
      isHeaderFixed,
      parallaxHeight,
      useNativeDriver,
      headerFixedTransformY,
      headerBackgroundColor,
      headerFixedBackgroundColor,
    } = props;

    const wrapperStyle = {
      position: 'absolute',
      top: 0,
      width,
      height,
      zIndex: 2,
    };
    const style = {
      flex: 1,
      width,
      height,
    };

    if (!isHeaderFixed) {
      const parallaxHeightWithoutHeader = parallaxHeight - height;

      style.transform = [
        {
          translateY: scrollY.interpolate({
            inputRange: parallaxHeightWithoutHeader
              ? [0, parallaxHeight - height, parallaxHeight]
              : [0, parallaxHeight],
            outputRange: parallaxHeightWithoutHeader
              ? [0, 0, -height]
              : [0, -height],
            extrapolate: 'clamp',
          }),
        },
      ];
    } else if (headerFixedTransformY) {
      style.transform = [
        {
          translateY: scrollY.interpolate({
            inputRange: [0, headerFixedTransformY],
            outputRange: [0, -headerFixedTransformY],
            extrapolate: 'clamp',
          }),
        },
      ];
    }

    if (!useNativeDriver && headerBackgroundColor) {
      style.backgroundColor = scrollY.interpolate({
        inputRange: [0, parallaxHeight - height],
        outputRange: [headerBackgroundColor, headerFixedBackgroundColor],
        extrapolate: 'clamp',
      });
    } else if (useNativeDriver && headerBackgroundColor) {
      style.backgroundColor = headerBackgroundColor;
    }

    return (
      <View style={wrapperStyle} pointerEvents="box-none">
        <Animated.View style={style} pointerEvents="box-none">
          {renderHeader({width, height, animatedValue: scrollY})}
        </Animated.View>
      </View>
    );
  };

  return (
    <View style={[wrapperStyle, {width, height}]}>
      {renderParallaxBackground && _renderParallaxBackground(!!renderHeader)}
      <ScrollableComponent
        {...scrollViewProps}
        ref={_ref}
        data={data && renderItem && _getFlatData(data)}
        style={style}
        throttle={16}
        onScroll={_onAnimatedScroll}
        sections={sections && _getSectionData(sections)}
        renderRow={dataSource && renderRow && _renderRow}
        renderItem={renderItem && _renderItem}
        dataSource={dataSource && renderRow && _getDataSource(dataSource)}
        scrollEventThrottle={16}
        stickyHeaderIndices={stickyHeaderIndices}
        contentContainerStyle={[
          contentContainerStyle,
          {width, minHeight: height},
        ]}
        stickySectionHeadersEnabled
        onMoveShouldSetResponder={() => false}
        bounces={false}
        alwaysBounceVertical={false}
        showsVerticalScrollIndicator={false}
        onStartShouldSetResponderCapture={() => false}>
        {isRenderChildComponents &&
          renderParallaxForeground &&
          _renderParallaxForeground()}
        {isRenderChildComponents && _renderBackgroundPlaceholder()}
        {isRenderChildComponents && children}
      </ScrollableComponent>
      {!isRenderChildComponents &&
        renderParallaxForeground &&
        _renderParallaxForeground()}
      {renderHeader && _renderHeader()}
    </View>
  );
};

export default ParallaxScroll;

ParallaxScroll.propTypes = {
  data: PropTypes.oneOfType([PropTypes.array]),
  style: PropTypes.oneOfType([PropTypes.number, PropTypes.object]),
  width: PropTypes.number,
  height: PropTypes.number,
  innerRef: PropTypes.func,
  sections: PropTypes.oneOfType([PropTypes.array]),
  onScroll: PropTypes.func,
  children: PropTypes.oneOfType([PropTypes.array, PropTypes.element]),
  renderRow: PropTypes.func,
  renderItem: PropTypes.func,
  dataSource: PropTypes.instanceOf(ListViewDataSource),
  scrollStyle: PropTypes.oneOfType([PropTypes.number, PropTypes.object]),
  headerHeight: PropTypes.number,
  renderHeader: PropTypes.func,
  isHeaderFixed: PropTypes.bool,
  onHeaderFixed: PropTypes.func,
  parallaxHeight: PropTypes.number,
  useNativeDriver: PropTypes.bool,
  backgroundScale: PropTypes.number,
  scrollableComponent: PropTypes.oneOfType([PropTypes.func, PropTypes.object]),
  isBackgroundScalable: PropTypes.bool,
  backgroundScaleOrigin: PropTypes.oneOf(['top', 'center']),
  headerFixedTransformY: PropTypes.number,
  headerBackgroundColor: PropTypes.string,
  contentContainerStyle: PropTypes.oneOfType([
    PropTypes.array,
    PropTypes.number,
    PropTypes.object,
  ]),
  onChangeHeaderVisibility: PropTypes.func,
  renderParallaxBackground: PropTypes.func,
  renderParallaxForeground: PropTypes.func,
  fadeOutParallaxForeground: PropTypes.bool,
  fadeOutParallaxBackground: PropTypes.bool,
  headerFixedBackgroundColor: PropTypes.string,
  renderBackgroundPlaceholder: PropTypes.func,
  parallaxBackgroundScrollSpeed: PropTypes.number,
  parallaxForegroundScrollSpeed: PropTypes.number,
};

ParallaxScroll.defaultProps = {
  data: [],
  style: {},
  width: width,
  height: height,
  innerRef: null,
  sections: null,
  children: null,
  onScroll: null,
  renderRow: null,
  dataSource: null,
  renderItem: null,
  scrollStyle: {},
  headerHeight: 45,
  renderHeader: null,
  isHeaderFixed: false,
  onHeaderFixed: () => {},
  parallaxHeight: width * RATIO,
  backgroundScale: 3,
  useNativeDriver: false,
  scrollableComponent: Animated.ScrollView,
  isBackgroundScalable: true,
  backgroundScaleOrigin: 'center',
  headerFixedTransformY: 0,
  headerBackgroundColor: 'rgba(0, 0, 0, 0)',
  contentContainerStyle: {},
  onChangeHeaderVisibility: () => {},
  renderParallaxBackground: null,
  renderParallaxForeground: null,
  fadeOutParallaxForeground: false,
  fadeOutParallaxBackground: false,
  headerFixedBackgroundColor: 'rgba(0, 0, 0, 1)',
  renderBackgroundPlaceholder: null,
  parallaxBackgroundScrollSpeed: 5,
  parallaxForegroundScrollSpeed: 5,
};
